# Singular Token Flags
NOT = "!"
QUOTED = "q"
UNQUOTED = "u"
CASE_SENSITIVE = "s"
CASE_INSENSITIVE = "i"

# Named Section Flags
ZERO_OR_ONE = "?"
ZERO_OR_MORE = "*"
ONE_OR_MORE = "+"

__MUTUALLY_EXCLUSIVE__ = (
    {CASE_SENSITIVE, CASE_INSENSITIVE},
    {QUOTED, UNQUOTED}
)

DEFAULTS = frozenset((
    CASE_INSENSITIVE,
))

__all__ = [
    "NOT",
    "QUOTED",
    "UNQUOTED",
    "CASE_SENSITIVE",
    "CASE_INSENSITIVE"
]



import _test_case
import tokex
from tokex.tokex_class import Tokex

class TestTokex(_test_case.TokexTestCase):

    def test_tokex_match(self):
        self.assertIsNotNone(tokex.match('"a" "b" "c"', 'a b c'))
        self.assertIsNone(tokex.match('"a" "b" "c"', 'a b c d'))
        self.assertIsNotNone(tokex.match('"a" "b" "c"', 'a b c d', match_entirety=False))
        self.assertIsNone(tokex.match('"a" "b" "c"', 'a', match_entirety=False))

        parser2 = tokex.compile("'a' def b{ 'b' } b()")

        self.assertIsNotNone(tokex.match("'a' def b{ 'b' } b()", 'a b'))
        self.assertIsNone(tokex.match("'a' def b { 'b' } b()", 'a b c'))
        self.assertIsNotNone(tokex.match("'a' def b { 'b' } b()", 'a b c', match_entirety=False))
        self.assertIsNone(tokex.match("'a' def b{ 'b' } b()", 'a', match_entirety=False))


    def test_tokex_compile(self):
        parser = tokex.compile('"a" "b" "c"')

        self.assertIsInstance(parser, Tokex)

        self.assertIsNotNone(parser.match('a b c'))
        self.assertIsNone(parser.match('a b c d'))
        self.assertIsNotNone(parser.match('a b c d', match_entirety=False))
        self.assertIsNone(parser.match('a', match_entirety=False))

        parser2 = tokex.compile("'a' def b{ 'b' } b()")

        self.assertIsInstance(parser2, Tokex)
        self.assertIsNotNone(parser2.match('a b'))
        self.assertIsNone(parser2.match('a b c'))
        self.assertIsNotNone(parser2.match('a b c', match_entirety=False))
        self.assertIsNone(parser2.match('a', match_entirety=False))



import re

from ...logger import LOGGER
from .. import flags

class BaseElement(object):
    """ Base class which all defined grammar element subclass from """

    # A set of flags which are valid to be set for this element
    valid_flags = None

    def __init__(self, token_str="", _flags=None, default_flags=flags.DEFAULTS):
        self.token_str = token_str
        # Records what flags were defined on the grammar
        self._grammar_flags = _flags
        # Records what flags will be used with the element; combines grammar flags with default flags
        self._flags = set(_flags) if _flags else None

        default_flags = default_flags or flags.DEFAULTS

        # Check that 2+ mutually exclusive flags weren't given
        if self._flags:
            for m_ex_set in flags.__MUTUALLY_EXCLUSIVE__:
                if len(m_ex_set.difference(self._flags)) < len(m_ex_set) - 1:
                    raise Exception("Mutually exclusive flags given: %s" % ", ".join(m_ex_set))

        # Set default flags
        if self.valid_flags and default_flags:
            for flag in default_flags:
                # Check if the default flag appears in any mutually exclusive set
                for m_ex_set in flags.__MUTUALLY_EXCLUSIVE__:
                    if flag in m_ex_set:
                        # If it does, check if we were given any of the other alternatives.  If not, apply the default
                        if len(m_ex_set.difference(self._flags or [])) == len(m_ex_set):
                            self._flags = self._flags or set()
                            self._flags.add(flag)

                        break

                # If we're in the else, the flag wasn't in any mutually exclusive set.  Apply it
                else:
                    self._flags = self._flags or set()
                    self._flags.add(flag)

        self.setup()

    def __repr__(self):
        return "<[%s]>" % self.human_readable_name()

    def human_readable_name(self):
        """ Returns a string which can be displayed to users, showing what sort of element they're looking at """

        return "Tokex Element"

    def _apply(self, string_tokens, idx):
        """
        Function which accepts an iterable of tokens and a current index and determines whether or not this token
        matches the list at the current position. Should be overridden in subclasses.

        Inputs: string_tokens - An iterable of string tokens to determine if we match upon.
                idx          - The start index within the string_tokens to begin processing at.

        Outputs: A triple containing: {
            match: A boolean depicting whether or not this construct matches the iterable at the current position.
            new_idx: The index this construct ceased matching upon the iterable, if match is True. Else None
            output: If there are any named matches within this construct, a dictionary or list.  Else None
        )
        """

        raise NotImplementedError

    def setup(self):
        pass

    def apply(self, string_tokens, idx):
        """
        Used to apply this token to an iterable of tokens at a specified position.  Uses self._apply to do the work
        of matching the inputs.

        Inputs: string_tokens - An iterable of string tokens to determine if we match upon.
                idx           - The start index within the string_tokens to begin processing at.

        Outputs: A triple containing: {
            match: A boolean depicting whether or not this construct matches the iterable at the current position.
            new_idx: The index this construct ceased matching upon the iterable, if match is True. Else None
            output: If there are any named matches within this construct, a dictionary or list.  Else None
        )
        """

        if idx < len(string_tokens):
            LOGGER.debug("%s testing match for: %s\n", self, string_tokens[idx])

        match, idx, output = self._apply(string_tokens, idx)

        if idx is not None and idx < len(string_tokens):
            LOGGER.debug("%s Matched: %s\n", self, match)

        return match, idx, output

    def has_flag(self, flag):
        """
        Returns a boolean indicating whether or not this grammar element was initialized with the given flag
        or, if not, if the flag was set as a default
        """

        return flag in (self._flags or '')


class BaseScopedElement(BaseElement):
    """ Base class for grammar elements which can have sub elements within them """

    name = None

    # Regular expression string which matches valid token names (ex: named sub grammars, named tokens, etc)
    name_re_str = "[a-zA-Z0-9_-]+"
    name_re = re.compile(name_re_str)

    def __init__(self, *args, **kwargs):
        super(BaseScopedElement, self).__init__(*args, **kwargs)

        # List of sub elements that this grammar element contains
        self.sub_elements = []

    def add_sub_element(self, sub_element):
        """
        Adds a sub element to this grammar element.  This is used in places such as zero or more for example,
        where the zero or more element will contain one or more sub elements

        Inputs: sub_element - The element to add as a sub element of this element.
                              Should be a subclass of BaseGrammar
        """

        self.sub_elements.append(sub_element)



import re

from tokex import tokenizers
import _test_case

class TestTokenizer(_test_case.TokexTestCase):

    input_string = """
        separate words which should be tokenized _separate_ly_
        (words) [to] {be} <tokenized> #separately# $from$ ;their; |surrounding| !characters!
    \n \t\n\n
        == <= >= != ... !=...<=
    """

    def test_tokenizer(self):
        tokenizer = tokenizers.TokexTokenizer()

        self.assertEqual(tokenizer.tokenize(self.input_string), [
            "separate", "words", "which", "should", "be", "tokenized", "_separate_ly_",
            "(", "words", ")", "[", "to", "]", "{", "be", "}", "<", "tokenized", ">", "#", "separately", "#",
            "$", "from", "$", ";", "their", ";", "|", "surrounding", "|", "!", "characters", "!",
            "==", "<=", ">=", "!=", "...", "!=...<="

        ])

    def test_tokenize_newlines(self):
        tokenizer = tokenizers.TokexTokenizer(tokenize_newlines=True)

        self.assertEqual(tokenizer.tokenize(self.input_string), [
            "\n", "separate", "words", "which", "should", "be", "tokenized", "_separate_ly_", "\n",
            "(", "words", ")", "[", "to", "]", "{", "be", "}", "<", "tokenized", ">", "#", "separately", "#",
            "$", "from", "$", ";", "their", ";", "|", "surrounding", "|", "!", "characters", "!", "\n", "\n", "\n", "\n", "\n",
            "==", "<=", ">=", "!=", "...", "!=...<=", "\n"
        ])

        tokenizer = tokenizers.TokexTokenizer(tokenize_newlines=True, ignore_empty_lines=True)

        self.assertEqual(tokenizer.tokenize(self.input_string), [
            "separate", "words", "which", "should", "be", "tokenized", "_separate_ly_", "\n",
            "(", "words", ")", "[", "to", "]", "{", "be", "}", "<", "tokenized", ">", "#", "separately", "#",
            "$", "from", "$", ";", "their", ";", "|", "surrounding", "|", "!", "characters", "!", "\n",
            "==", "<=", ">=", "!=", "...", "!=...<=", "\n"
        ])


class TestNumericTokenizer(_test_case.TokexTestCase):

    input_string = r"""
        separate words which should be tokenized _separate_ly_

        $1230.1230 123. 12430$ 123,123 1234 123.123

        == <= >= != ... !=...<=
    """

    def test_numeric_tokenizer(self):
        tokenizer = tokenizers.NumericTokenizer()

        self.assertEqual(tokenizer.tokenize(self.input_string), [
            "separate", "words", "which", "should", "be", "tokenized", "_separate_ly_",

            "$1230.1230", "123.", "12430$", "123,123", "1234", "123.123",

            "==", "<=", ">=", "!=", "...", "!=...<="
        ])



from .grammar import flags
from .tokex_class import Tokex
from .tokenizers.tokenizer import TokexTokenizer


def compile(input_grammar,
            allow_sub_grammar_definitions=True,
            tokenizer=TokexTokenizer,
            default_flags=flags.DEFAULTS):
    """
    Constructs and returns an instance of _StringParser for repeated parsing of strings using the given grammar.

    Inputs: input_grammar - The grammar to use to parse the input string.
            allow_sub_grammar_definitions - A Boolean, indicating whether or not sub grammar declarations,
                                            (@name: grammar @@), should be processed.  If grammarString has
                                            come from an untrusted source this should be set to False, to
                                            mitigate the potential for a `Billion Laughs` attack.
            tokenizer - Optional: An instance or class of tokenizers.Tokenizer.TokexTokenizer or a subclass of it,
                        used to tokenize the input string for parsing. Defaults to the base class TokexTokenizer.
            default_flags - A set of flags which will apply to all elements by default.
                            Default flags can be overridden by specifying an opposing flag on elements in the grammar.


    Outputs: An instance of _StringParser whose `match` function can be used to repeatedly parse input strings.
    """

    return Tokex(input_grammar, allow_sub_grammar_definitions, tokenizer, default_flags=default_flags)


def match(input_grammar,
          input_string,
          match_entirety=True,
          allow_sub_grammar_definitions=True,
          tokenizer=TokexTokenizer,
          default_flags=flags.DEFAULTS,
          debug=False):
    """
    Convenience function for performing matches using a grammar against a string.

    Inputs: input_grammar  - The grammar to use to parse the input string.
            input_string   - The string to be parsed.
            match_entirety - A boolean, if True requires the entire string to be matched by the grammar.
                             if False, trailing tokens not matched by the grammar will not cause a match failure.
            allow_sub_grammar_definitions - A Boolean, indicating whether or not sub grammar declarations,
                                         (@name: grammar @@), should be processed.  If grammarString has
                                         come from an untrusted source this should be set to False, to
                                         mitigate the potential for a `Billion Laughs` attack.
            tokenizer      - Optional: An instance or class of tokenizers.Tokenizer.TokexTokenizer or a subclass of it,
                             used to tokenize the input string for parsing. Defaults to the base class TokexTokenizer.
            default_flags - A set of flags which will apply to all elements by default.
                            Default flags can be overridden by specifying an opposing flag on elements in the grammar.
            debug          - A boolean, if True will set the debugging level to DEBUG for the duration of the match

    Outputs: The result of matching the input_string, if it matches, else None.
    """

    return Tokex(
        input_grammar,
        allow_sub_grammar_definitions,
        tokenizer,
        default_flags=default_flags,
    ).match(input_string, match_entirety=match_entirety, debug=debug)



import re

from ... import errors
from .. import flags

from ._base_element import BaseElement

class BaseSingular(BaseElement):
    """ Base class for singular elements, as most share a similar process flow """

    def _apply_first(self, string_tokens, idx):
        """
        Function which performs initial checking of the string/tokens idx

        Outputs: If successful, a string containing the token to match.
                 If unsuccessful, None
        """

        # If the index we're considering is beyond the end of our tokens, we have nothing to match on. Return False.
        if idx >= len(string_tokens):
            return None

        to_match = string_tokens[idx]

        if self.has_flag(flags.CASE_INSENSITIVE):
            to_match = to_match.lower()

        is_quoted = to_match[0] in ('"', "'") and to_match[-1] == to_match[0]

        if self.has_flag(flags.QUOTED):
            if is_quoted:
                return to_match[1:-1]

            return None

        if self.has_flag(flags.UNQUOTED) and is_quoted:
            return None

        return to_match


class AnyString(BaseSingular):
    valid_flags = {
        flags.QUOTED,
        flags.UNQUOTED
    }

    def human_readable_name(self):
        return "Any String  ."

    def _apply(self, string_tokens, idx):
        if self._apply_first(string_tokens, idx) is not None:
            return True, idx + 1, None

        return False, None, None


class Newline(BaseSingular):
    def human_readable_name(self):
        return "Newline $"

    def _apply(self, string_tokens, idx):
        to_match = self._apply_first(string_tokens, idx)

        if to_match == "\n":
            return True, idx + 1, None

        return False, None, None


class StringLiteral(BaseSingular):
    valid_flags = {
        flags.CASE_SENSITIVE,
        flags.CASE_INSENSITIVE,
        flags.QUOTED,
        flags.UNQUOTED,
        flags.NOT
    }

    def setup(self):
        # Strip the '/" away
        self.token_str = self.token_str[1:-1]

        if self.has_flag(flags.CASE_INSENSITIVE):
            self.token_str = self.token_str.lower()

    def human_readable_name(self):
        return "String Literal %s" % self.token_str

    def _apply(self, string_tokens, idx):
        to_match = self._apply_first(string_tokens, idx)

        if to_match is not None:
            matches = (to_match == self.token_str)
            if matches ^ self.has_flag(flags.NOT):
                return True, idx + 1, None

        return False, None, None

class RegexString(BaseSingular):
    valid_flags = {
        flags.CASE_SENSITIVE,
        flags.CASE_INSENSITIVE,
        flags.QUOTED,
        flags.UNQUOTED,
        flags.NOT
    }

    def setup(self):
        # Strip the ~ away
        self.token_str = self.token_str[1:-1]

        if self.has_flag(flags.CASE_INSENSITIVE):
            self.token_str = self.token_str.lower()

        try:
            self.regex = re.compile(self.token_str)

        except re.error:
            raise errors.GrammarParsingError("Invalid Regular Expression given: %s" % self.token_str)

    def human_readable_name(self):
        return "Regular Expression %s" % self.token_str

    def _apply(self, string_tokens, idx):
        to_match = self._apply_first(string_tokens, idx)

        if to_match is not None:
            matches = bool(self.regex.match(to_match))

            if matches ^ self.has_flag(flags.NOT):
                return True, idx + 1, None

        return False, None, None



import re

from tokex.grammar.parse import tokenize_grammar, construct_grammar
from tokex.grammar import elements
from tokex.grammar import flags
from tokex import errors
import _test_case

class TestGrammarConstruction(_test_case.TokexTestCase):
    """ Class which tests the construction of a Tokex grammar from a grammar string """

    def test_tokenize_grammar(self):
        grammar_string = r"""
            { 'a' 'b' 'c' }
            ?(a: 'd' 'e' 'f' )
            *(b
: 'g' 'h' 'i' )
            +(1 : 'g' 'h' 'i' )
            ?(ab_12:'d' 'e' 'f' )
            ?('d' 'e' 'f' )
            *(a:'g' 'h' 'i' )
            +(a: 'g' 'h' 'i' )
            *(a:'g' 'h' 'i' sep {'a'})
            +(a: 'g' 'h' 'i' sep{ (b: 'b') })
            ( abc: 'abc' '123')
            (
             def: 'def' '456' )
            (ghi: 'ghi' '789')
            < abc: '123'>
            <
             def: '456' >
            <ghi: '789'>
            def abc { '123'}
            def
                def{
             '456' }
            def    ghi
                { '789'
                }

            $

            ~someRegex~
            ~ someRegex~
            ~someRegex ~
            ~ someRegex ~
        """

        grammar_tokens = [t["token"] for t in tokenize_grammar(grammar_string)]

        self.assertEqual(grammar_tokens, [
            "{", "'a'", "'b'", "'c'", "}",
            "?(a:", "'d'", "'e'", "'f'", ")",
            "*(b\n:", "'g'", "'h'", "'i'", ")",
            "+(1 :", "'g'", "'h'", "'i'", ")",
            "?(ab_12:", "'d'", "'e'", "'f'", ")",
            "?(", "'d'", "'e'", "'f'", ")",
            "*(a:", "'g'", "'h'", "'i'", ")",
            "+(a:", "'g'", "'h'", "'i'", ")",
            "*(a:", "'g'", "'h'", "'i'", "sep {", "'a'", "}", ")",
            "+(a:", "'g'", "'h'", "'i'", "sep{", "(b:", "'b'", ")", "}", ")",
            "( abc:", "'abc'", "'123'", ")",
            "(\n             def:", "'def'", "'456'", ")",
            "(ghi:", "'ghi'", "'789'", ")",
            "< abc:", "'123'", ">",
            "<\n             def:", "'456'", ">",
            "<ghi:", "'789'", ">",
            "def abc {", "'123'", "}",
            "def\n                def{", "'456'", "}",
            "def    ghi\n                {", "'789'", "}",
            "$",
            "~someRegex~", "~ someRegex~", "~someRegex ~", "~ someRegex ~",
        ])

        flags_grammar_string = r"""
            !'abc'
            !" abc"
            !"abc \" def"
            u.
            q.
            .
            s~AnotherRegex~
            s~AnotherRegex ~
            s~ AnotherRegex~
            s~ AnotherRegex ~
            "string"
            " string "
            "str ing"
            "str'ing"
            "str\"ing"
            "str\"ing\\"
            ""
            'string'
            ' string '
            'str ing'
            'str"ing'
            'str\'ing'
            ''
            s"string"
            s" string "
            s"str ing"
            s"str'ing"
            s"str\"ing"
            s"str\`ing"
            s""
            i"string"
            i" string "
            i"str ing"
            i"str'ing"
            i"str\"ing"
            i"str\`ing"
            i""
        """

        grammar_tokens = ["%s|%s" % ("".join(sorted(t["flags"] or [])), t["token"]) for t in tokenize_grammar(flags_grammar_string)]

        self.assertEqual(grammar_tokens, [
            "!|'abc'", '!|" abc"', '!|"abc " def"',
            "u|.", "q|.", "|.",
            "s|~AnotherRegex~", "s|~AnotherRegex ~", "s|~ AnotherRegex~", "s|~ AnotherRegex ~",
            '|"string"', '|" string "', '|"str ing"', '|"str\'ing"', '|"str"ing"', '|"str"ing\\"', '|""',
            "|'string'", "|' string '", "|'str ing'", "|'str\"ing'", "|'str'ing'", "|''",
            "s|\"string\"", "s|\" string \"", "s|\"str ing\"", "s|\"str'ing\"", 's|\"str"ing\"', "s|\"str`ing\"", "s|\"\"",
            "i|\"string\"", "i|\" string \"", "i|\"str ing\"", "i|\"str'ing\"", 'i|\"str"ing\"', "i|\"str`ing\"", "i|\"\"",
        ])

    def test_parse_token(self):
        test_grammar = construct_grammar(r"""
            "a" "a b" 'c' 'efg' s'h' s"i j k" ~l~ ~m n o~ s~p~ s~q r s~
            "a\"'`$~b\\" 'c\'"`$^d\\\\' u~g`'"\~$h~ q~i`'"\~\\^j\\\\~
            .
            q.
            u.
            !"a`'\"$^"
            $
        """)

        self.assertIsInstance(test_grammar.sub_elements[0], elements.StringLiteral)
        self.assertEqual(test_grammar.sub_elements[0].token_str, 'a')
        self.assertIsNone(test_grammar.sub_elements[0]._grammar_flags)

        self.assertIsInstance(test_grammar.sub_elements[1], elements.StringLiteral)
        self.assertEqual(test_grammar.sub_elements[1].token_str, 'a b')
        self.assertIsNone(test_grammar.sub_elements[1]._grammar_flags)

        self.assertIsInstance(test_grammar.sub_elements[2], elements.StringLiteral)
        self.assertEqual(test_grammar.sub_elements[2].token_str, 'c')
        self.assertIsNone(test_grammar.sub_elements[2]._grammar_flags)

        self.assertIsInstance(test_grammar.sub_elements[3], elements.StringLiteral)
        self.assertEqual(test_grammar.sub_elements[3].token_str, 'efg')
        self.assertIsNone(test_grammar.sub_elements[3]._grammar_flags)

        self.assertIsInstance(test_grammar.sub_elements[4], elements.StringLiteral)
        self.assertEqual(test_grammar.sub_elements[4].token_str, 'h')
        self.assertEqual(test_grammar.sub_elements[4]._grammar_flags, {flags.CASE_SENSITIVE})

        self.assertIsInstance(test_grammar.sub_elements[5], elements.StringLiteral)
        self.assertEqual(test_grammar.sub_elements[5].token_str, 'i j k')
        self.assertEqual(test_grammar.sub_elements[5]._grammar_flags, {flags.CASE_SENSITIVE})

        self.assertIsInstance(test_grammar.sub_elements[6], elements.RegexString)
        self.assertEqual(test_grammar.sub_elements[6].token_str, 'l')
        self.assertIsNone(test_grammar.sub_elements[6]._grammar_flags)

        self.assertIsInstance(test_grammar.sub_elements[7], elements.RegexString)
        self.assertEqual(test_grammar.sub_elements[7].token_str, 'm n o')
        self.assertIsNone(test_grammar.sub_elements[7]._grammar_flags)

        self.assertIsInstance(test_grammar.sub_elements[8], elements.RegexString)
        self.assertEqual(test_grammar.sub_elements[8].token_str, 'p')
        self.assertEqual(test_grammar.sub_elements[8]._grammar_flags, {flags.CASE_SENSITIVE})

        self.assertIsInstance(test_grammar.sub_elements[9], elements.RegexString)
        self.assertEqual(test_grammar.sub_elements[9].token_str, 'q r s')
        self.assertEqual(test_grammar.sub_elements[9]._grammar_flags, {flags.CASE_SENSITIVE})

        self.assertIsInstance(test_grammar.sub_elements[10], elements.StringLiteral)
        self.assertEqual(test_grammar.sub_elements[10].token_str, 'a"\'`$~b\\')
        self.assertIsNone(test_grammar.sub_elements[10]._grammar_flags)

        self.assertIsInstance(test_grammar.sub_elements[11], elements.StringLiteral)
        self.assertEqual(test_grammar.sub_elements[11].token_str, 'c\'"`$^d\\\\')
        self.assertIsNone(test_grammar.sub_elements[11]._grammar_flags)

        self.assertIsInstance(test_grammar.sub_elements[12], elements.RegexString)
        self.assertEqual(test_grammar.sub_elements[12].token_str, 'g`\'"~$h')
        self.assertEqual(test_grammar.sub_elements[12]._grammar_flags, {flags.UNQUOTED})

        self.assertIsInstance(test_grammar.sub_elements[13], elements.RegexString)
        self.assertEqual(test_grammar.sub_elements[13].token_str, 'i`\'"~\\^j\\\\')
        self.assertEqual(test_grammar.sub_elements[13]._grammar_flags, {flags.QUOTED})

        self.assertIsInstance(test_grammar.sub_elements[14], elements.AnyString)
        self.assertIsNone(test_grammar.sub_elements[14]._grammar_flags)

        self.assertIsInstance(test_grammar.sub_elements[15], elements.AnyString)
        self.assertEqual(test_grammar.sub_elements[15]._grammar_flags, {flags.QUOTED})

        self.assertIsInstance(test_grammar.sub_elements[16], elements.AnyString)
        self.assertEqual(test_grammar.sub_elements[16]._grammar_flags, {flags.UNQUOTED})

        self.assertIsInstance(test_grammar.sub_elements[17], elements.StringLiteral)
        self.assertEqual(test_grammar.sub_elements[17].token_str, 'a`\'"$^')
        self.assertEqual(test_grammar.sub_elements[17]._grammar_flags, {flags.NOT})

        self.assertIsInstance(test_grammar.sub_elements[18], elements.Newline)
        self.assertIsNone(test_grammar.sub_elements[18]._grammar_flags)

    def test_parse_comment(self):
        test_grammar = construct_grammar(r"""
            s'a' i'b' 'c' # letters
            # some more letters
            q'q' u"r" !'e'
        """)

        self.assertListEqual(
            [t.token_str for t in test_grammar.sub_elements],
            ['a', 'b', 'c', 'q', 'r', 'e']
        )

        test_grammar = construct_grammar("'c' # letters")

        self.assertListEqual(
            [t.token_str for t in test_grammar.sub_elements],
            ['c']
        )

    def test_parse_error(self):
        self.assertRaises(errors.GrammarParsingError, construct_grammar, "a")
        self.assertRaises(errors.GrammarParsingError, construct_grammar, "()")
        self.assertRaises(errors.GrammarParsingError, construct_grammar, ",")
        self.assertRaises(errors.GrammarParsingError, construct_grammar, "<>")
        self.assertRaises(errors.GrammarParsingError, construct_grammar, "[]")
        self.assertRaises(errors.GrammarParsingError, construct_grammar, "(asdf asdf : 'a')")
        self.assertRaises(errors.GrammarParsingError, construct_grammar, "'a\"")
        self.assertRaises(errors.GrammarParsingError, construct_grammar, r"'a\'")

    def test_parse_named_element(self):
        test_grammar = construct_grammar(r"""
            <abc: "123">
            <def : !'456'>
            < ghi : s~7<>8~>
            <
             jkl: q.>
            <mno: $>
        """)

        self.assertIsInstance(test_grammar.sub_elements[0], elements.NamedElement)
        self.assertEqual(test_grammar.sub_elements[0].name, 'abc')
        self.assertIsInstance(test_grammar.sub_elements[0].sub_elements[0], elements.StringLiteral)
        self.assertEqual(test_grammar.sub_elements[0].sub_elements[0].token_str, "123")
        self.assertIsNone(test_grammar.sub_elements[0].sub_elements[0]._grammar_flags)

        self.assertIsInstance(test_grammar.sub_elements[1], elements.NamedElement)
        self.assertEqual(test_grammar.sub_elements[1].name, 'def')
        self.assertIsInstance(test_grammar.sub_elements[1].sub_elements[0], elements.StringLiteral)
        self.assertEqual(test_grammar.sub_elements[1].sub_elements[0].token_str, "456")
        self.assertEqual(test_grammar.sub_elements[1].sub_elements[0]._grammar_flags, {flags.NOT})

        self.assertIsInstance(test_grammar.sub_elements[2], elements.NamedElement)
        self.assertEqual(test_grammar.sub_elements[2].name, 'ghi')
        self.assertIsInstance(test_grammar.sub_elements[2].sub_elements[0], elements.RegexString)
        self.assertEqual(test_grammar.sub_elements[2].sub_elements[0].token_str, "7<>8")
        self.assertEqual(test_grammar.sub_elements[2].sub_elements[0]._grammar_flags, {flags.CASE_SENSITIVE})

        self.assertIsInstance(test_grammar.sub_elements[3], elements.NamedElement)
        self.assertEqual(test_grammar.sub_elements[3].name, 'jkl')
        self.assertIsInstance(test_grammar.sub_elements[3].sub_elements[0], elements.AnyString)
        self.assertEqual(test_grammar.sub_elements[3].sub_elements[0]._grammar_flags, {flags.QUOTED})

        self.assertIsInstance(test_grammar.sub_elements[4], elements.NamedElement)
        self.assertEqual(test_grammar.sub_elements[4].name, 'mno')
        self.assertIsInstance(test_grammar.sub_elements[4].sub_elements[0], elements.Newline)
        self.assertIsNone(test_grammar.sub_elements[4].sub_elements[0]._grammar_flags)

        self.assertRaises(errors.GrammarParsingError, construct_grammar, "<mno: 'a' 'b'>")
        self.assertRaises(errors.GrammarParsingError, construct_grammar, "<mno: {'a' 'b'}>")
        self.assertRaises(errors.GrammarParsingError, construct_grammar, "<mno: 'a'")
        self.assertRaises(errors.GrammarParsingError, construct_grammar, "mno: 'a'>")
        self.assertRaises(errors.GrammarParsingError, construct_grammar, '<"mno": "a">')
        self.assertRaises(errors.GrammarParsingError, construct_grammar, "<'mno': 'a'>")
        self.assertRaises(errors.GrammarParsingError, construct_grammar, "<a: <b: 'a'>>")


    def test_parse_grammar(self):
        test_grammar = construct_grammar(r"""
            (abc: "123")
            (def :q'456')
            ( ghi : s'7()8')
            (
             jkl: !~()~)
            (mno: (pqr: .) )
            (stu: i'a' $ q"c")
        """)

        self.assertIsInstance(test_grammar.sub_elements[0], elements.Grammar)
        self.assertEqual(test_grammar.sub_elements[0].name, 'abc')
        self.assertIsInstance(test_grammar.sub_elements[0].sub_elements[0], elements.StringLiteral)
        self.assertEqual(test_grammar.sub_elements[0].sub_elements[0].token_str, '123')
        self.assertIsNone(test_grammar.sub_elements[0].sub_elements[0]._grammar_flags)

        self.assertIsInstance(test_grammar.sub_elements[1], elements.Grammar)
        self.assertEqual(test_grammar.sub_elements[1].name, 'def')
        self.assertIsInstance(test_grammar.sub_elements[1].sub_elements[0], elements.StringLiteral)
        self.assertEqual(test_grammar.sub_elements[1].sub_elements[0].token_str, '456')
        self.assertEqual(test_grammar.sub_elements[1].sub_elements[0]._grammar_flags, {flags.QUOTED})

        self.assertIsInstance(test_grammar.sub_elements[2], elements.Grammar)
        self.assertEqual(test_grammar.sub_elements[2].name, 'ghi')
        self.assertIsInstance(test_grammar.sub_elements[2].sub_elements[0], elements.StringLiteral)
        self.assertEqual(test_grammar.sub_elements[2].sub_elements[0].token_str, '7()8')
        self.assertEqual(test_grammar.sub_elements[2].sub_elements[0]._grammar_flags, {flags.CASE_SENSITIVE})

        self.assertIsInstance(test_grammar.sub_elements[3], elements.Grammar)
        self.assertEqual(test_grammar.sub_elements[3].name, 'jkl')
        self.assertIsInstance(test_grammar.sub_elements[3].sub_elements[0], elements.RegexString)
        self.assertEqual(test_grammar.sub_elements[3].sub_elements[0].token_str, '()')
        self.assertEqual(test_grammar.sub_elements[3].sub_elements[0]._grammar_flags, {flags.NOT})

        self.assertIsInstance(test_grammar.sub_elements[4], elements.Grammar)
        self.assertEqual(test_grammar.sub_elements[4].name, 'mno')
        self.assertIsInstance(test_grammar.sub_elements[4].sub_elements[0], elements.Grammar)
        self.assertEqual(test_grammar.sub_elements[4].sub_elements[0].name, 'pqr')
        self.assertIsInstance(test_grammar.sub_elements[4].sub_elements[0].sub_elements[0], elements.AnyString)
        self.assertIsNone(test_grammar.sub_elements[4].sub_elements[0].sub_elements[0]._grammar_flags)

        self.assertIsInstance(test_grammar.sub_elements[5], elements.Grammar)
        self.assertEqual(test_grammar.sub_elements[5].name, 'stu')
        self.assertIsInstance(test_grammar.sub_elements[5].sub_elements[0], elements.StringLiteral)
        self.assertEqual(test_grammar.sub_elements[5].sub_elements[0].token_str, 'a')
        self.assertEqual(test_grammar.sub_elements[5].sub_elements[0]._grammar_flags, {flags.CASE_INSENSITIVE})
        self.assertIsInstance(test_grammar.sub_elements[5].sub_elements[1], elements.Newline)
        self.assertIsNone(test_grammar.sub_elements[5].sub_elements[1]._grammar_flags)
        self.assertIsInstance(test_grammar.sub_elements[5].sub_elements[2], elements.StringLiteral)
        self.assertEqual(test_grammar.sub_elements[5].sub_elements[2].token_str, 'c')
        self.assertEqual(test_grammar.sub_elements[5].sub_elements[2]._grammar_flags, {flags.QUOTED})

        self.assertRaises(errors.GrammarParsingError, construct_grammar, "(mno: 'a'")
        self.assertRaises(errors.GrammarParsingError, construct_grammar, "(m no: 'a')")
        self.assertRaises(errors.GrammarParsingError, construct_grammar, "mno: 'a')")
        self.assertRaises(errors.GrammarParsingError, construct_grammar, '("mno": "a")')
        self.assertRaises(errors.GrammarParsingError, construct_grammar, "('mno': 'a')")
        self.assertRaises(errors.GrammarParsingError, construct_grammar, "('mno': 'a' ['b'])")


    def test_parse_zero_or_one(self):
        test_grammar = construct_grammar(r"""
            ?(a:
              ?(b_: 'a' !"b" q'c')
              ?(-c:. q. u. )
              ?(  de:{$ .} )
              ?(.)
              ?()
            )
        """)

        self.assertIsInstance(test_grammar.sub_elements[0], elements.ZeroOrOne)
        self.assertEqual(test_grammar.sub_elements[0].name, 'a')
        self.assertIsInstance(test_grammar.sub_elements[0].sub_elements[0], elements.ZeroOrOne)
        self.assertEqual(test_grammar.sub_elements[0].sub_elements[0].name, 'b_')
        self.assertIsInstance(test_grammar.sub_elements[0].sub_elements[1], elements.ZeroOrOne)
        self.assertEqual(test_grammar.sub_elements[0].sub_elements[1].name, '-c')
        self.assertIsInstance(test_grammar.sub_elements[0].sub_elements[2], elements.ZeroOrOne)
        self.assertEqual(test_grammar.sub_elements[0].sub_elements[2].name, 'de')
        self.assertIsInstance(test_grammar.sub_elements[0].sub_elements[3], elements.ZeroOrOne)
        self.assertIsNone(test_grammar.sub_elements[0].sub_elements[3].name)
        self.assertIsInstance(test_grammar.sub_elements[0].sub_elements[4], elements.ZeroOrOne)
        self.assertIsNone(test_grammar.sub_elements[0].sub_elements[4].name)

        se = test_grammar.sub_elements[0].sub_elements

        self.assertIsInstance(se[0].sub_elements[0], elements.StringLiteral)
        self.assertEqual(se[0].sub_elements[0].token_str, 'a')
        self.assertIsNone(se[0].sub_elements[0]._grammar_flags)

        self.assertIsInstance(se[0].sub_elements[1], elements.StringLiteral)
        self.assertEqual(se[0].sub_elements[1].token_str, 'b')
        self.assertEqual(se[0].sub_elements[1]._grammar_flags, {flags.NOT})

        self.assertIsInstance(se[0].sub_elements[2], elements.StringLiteral)
        self.assertEqual(se[0].sub_elements[2].token_str, 'c')
        self.assertEqual(se[0].sub_elements[2]._grammar_flags, {flags.QUOTED})


        self.assertIsInstance(se[1].sub_elements[0], elements.AnyString)
        self.assertIsNone(se[1].sub_elements[0]._grammar_flags)

        self.assertIsInstance(se[1].sub_elements[1], elements.AnyString)
        self.assertEqual(se[1].sub_elements[1]._grammar_flags, {flags.QUOTED})

        self.assertIsInstance(se[1].sub_elements[2], elements.AnyString)
        self.assertEqual(se[1].sub_elements[2]._grammar_flags, {flags.UNQUOTED})


        self.assertIsInstance(se[2].sub_elements[0], elements.OneOfSet)
        self.assertIsNone(se[2].sub_elements[0]._grammar_flags)

        self.assertIsInstance(se[2].sub_elements[0].sub_elements[0], elements.Newline)
        self.assertIsNone(se[2].sub_elements[0].sub_elements[0]._grammar_flags)

        self.assertIsInstance(se[2].sub_elements[0].sub_elements[1], elements.AnyString)
        self.assertIsNone(se[2].sub_elements[0].sub_elements[1]._grammar_flags)

        self.assertIsInstance(se[3].sub_elements[0], elements.AnyString)
        self.assertIsNone(se[3].sub_elements[0]._grammar_flags)

        self.assertEqual(len(se[4].sub_elements), 0)

        self.assertRaises(errors.GrammarParsingError, construct_grammar, "?(a:")
        self.assertRaises(errors.GrammarParsingError, construct_grammar, ")")
        self.assertRaises(errors.GrammarParsingError, construct_grammar, "?(a:))")
        self.assertRaises(errors.GrammarParsingError, construct_grammar, "?(a:{)}")
        self.assertRaises(errors.GrammarParsingError, construct_grammar, "? (a: 'a' )")
        self.assertRaises(errors.GrammarParsingError, construct_grammar, "?(a: 'a' sep { 'b'})")


    def test_parse_zero_or_more(self):
        test_grammar = construct_grammar(r"""
            *(a:
              *(b_: 'a' !"b" q'c')
              *(-c:. q. u. )
              *(  de:{$ .} sep { !~a~})
            )
        """)

        self.assertIsInstance(test_grammar.sub_elements[0], elements.ZeroOrMore)
        self.assertEqual(test_grammar.sub_elements[0].name, 'a')
        self.assertIsInstance(test_grammar.sub_elements[0].sub_elements[0], elements.ZeroOrMore)
        self.assertEqual(test_grammar.sub_elements[0].sub_elements[0].name, 'b_')
        self.assertIsInstance(test_grammar.sub_elements[0].sub_elements[1], elements.ZeroOrMore)
        self.assertEqual(test_grammar.sub_elements[0].sub_elements[1].name, '-c')
        self.assertIsInstance(test_grammar.sub_elements[0].sub_elements[2], elements.ZeroOrMore)
        self.assertEqual(test_grammar.sub_elements[0].sub_elements[2].name, 'de')

        se = test_grammar.sub_elements[0].sub_elements

        self.assertIsInstance(se[0].sub_elements[0], elements.StringLiteral)
        self.assertEqual(se[0].sub_elements[0].token_str, 'a')
        self.assertIsNone(se[0].sub_elements[0]._grammar_flags)

        self.assertIsInstance(se[0].sub_elements[1], elements.StringLiteral)
        self.assertEqual(se[0].sub_elements[1].token_str, 'b')
        self.assertEqual(se[0].sub_elements[1]._grammar_flags, {flags.NOT})

        self.assertIsInstance(se[0].sub_elements[2], elements.StringLiteral)
        self.assertEqual(se[0].sub_elements[2].token_str, 'c')
        self.assertEqual(se[0].sub_elements[2]._grammar_flags, {flags.QUOTED})


        self.assertIsInstance(se[1].sub_elements[0], elements.AnyString)
        self.assertIsNone(se[1].sub_elements[0]._grammar_flags)

        self.assertIsInstance(se[1].sub_elements[1], elements.AnyString)
        self.assertEqual(se[1].sub_elements[1]._grammar_flags, {flags.QUOTED})

        self.assertIsInstance(se[1].sub_elements[2], elements.AnyString)
        self.assertEqual(se[1].sub_elements[2]._grammar_flags, {flags.UNQUOTED})


        self.assertIsInstance(se[2].sub_elements[0], elements.OneOfSet)
        self.assertIsNone(se[2].sub_elements[0]._grammar_flags)

        self.assertIsInstance(se[2].sub_elements[0].sub_elements[0], elements.Newline)
        self.assertIsNone(se[2].sub_elements[0].sub_elements[0]._grammar_flags)

        self.assertIsInstance(se[2].sub_elements[0].sub_elements[1], elements.AnyString)
        self.assertIsNone(se[2].sub_elements[0].sub_elements[1]._grammar_flags)

        self.assertIsNotNone(se[2].delimiter_grammar)
        self.assertIsInstance(se[2].delimiter_grammar.sub_elements[0], elements.RegexString)
        self.assertEqual(se[2].delimiter_grammar.sub_elements[0].token_str, 'a')
        self.assertEqual(se[2].delimiter_grammar.sub_elements[0]._grammar_flags, {flags.NOT})

        self.assertRaises(errors.GrammarParsingError, construct_grammar, "*(a:")
        self.assertRaises(errors.GrammarParsingError, construct_grammar, ")")
        self.assertRaises(errors.GrammarParsingError, construct_grammar, "*(a:))")
        self.assertRaises(errors.GrammarParsingError, construct_grammar, "*(a:{)}")
        self.assertRaises(errors.GrammarParsingError, construct_grammar, "* (a: 'a' )")
        self.assertRaises(errors.GrammarParsingError, construct_grammar, "*(a: 'a' sep)")
        self.assertRaises(errors.GrammarParsingError, construct_grammar, "*(a: 'a' sep {)")


    def test_parse_one_or_more(self):
        test_grammar = construct_grammar(r"""
            +(a:
              +(b_: 'a' !"b" q'c')
              +(-c:. q. u. )
              +(  de:{$ .} sep { !~a~})
            )
        """)

        self.assertIsInstance(test_grammar.sub_elements[0], elements.OneOrMore)
        self.assertEqual(test_grammar.sub_elements[0].name, 'a')
        self.assertIsInstance(test_grammar.sub_elements[0].sub_elements[0], elements.OneOrMore)
        self.assertEqual(test_grammar.sub_elements[0].sub_elements[0].name, 'b_')
        self.assertIsInstance(test_grammar.sub_elements[0].sub_elements[1], elements.OneOrMore)
        self.assertEqual(test_grammar.sub_elements[0].sub_elements[1].name, '-c')
        self.assertIsInstance(test_grammar.sub_elements[0].sub_elements[2], elements.OneOrMore)
        self.assertEqual(test_grammar.sub_elements[0].sub_elements[2].name, 'de')

        se = test_grammar.sub_elements[0].sub_elements

        self.assertIsInstance(se[0].sub_elements[0], elements.StringLiteral)
        self.assertEqual(se[0].sub_elements[0].token_str, 'a')
        self.assertIsNone(se[0].sub_elements[0]._grammar_flags)

        self.assertIsInstance(se[0].sub_elements[1], elements.StringLiteral)
        self.assertEqual(se[0].sub_elements[1].token_str, 'b')
        self.assertEqual(se[0].sub_elements[1]._grammar_flags, {flags.NOT})

        self.assertIsInstance(se[0].sub_elements[2], elements.StringLiteral)
        self.assertEqual(se[0].sub_elements[2].token_str, 'c')
        self.assertEqual(se[0].sub_elements[2]._grammar_flags, {flags.QUOTED})


        self.assertIsInstance(se[1].sub_elements[0], elements.AnyString)
        self.assertIsNone(se[1].sub_elements[0]._grammar_flags)

        self.assertIsInstance(se[1].sub_elements[1], elements.AnyString)
        self.assertEqual(se[1].sub_elements[1]._grammar_flags, {flags.QUOTED})

        self.assertIsInstance(se[1].sub_elements[2], elements.AnyString)
        self.assertEqual(se[1].sub_elements[2]._grammar_flags, {flags.UNQUOTED})


        self.assertIsInstance(se[2].sub_elements[0], elements.OneOfSet)
        self.assertIsNone(se[2].sub_elements[0]._grammar_flags)

        self.assertIsInstance(se[2].sub_elements[0].sub_elements[0], elements.Newline)
        self.assertIsNone(se[2].sub_elements[0].sub_elements[0]._grammar_flags)

        self.assertIsInstance(se[2].sub_elements[0].sub_elements[1], elements.AnyString)
        self.assertIsNone(se[2].sub_elements[0].sub_elements[1]._grammar_flags)

        self.assertIsNotNone(se[2].delimiter_grammar)
        self.assertIsInstance(se[2].delimiter_grammar.sub_elements[0], elements.RegexString)
        self.assertEqual(se[2].delimiter_grammar.sub_elements[0].token_str, 'a')
        self.assertEqual(se[2].delimiter_grammar.sub_elements[0]._grammar_flags, {flags.NOT})

        self.assertRaises(errors.GrammarParsingError, construct_grammar, "+(a:")
        self.assertRaises(errors.GrammarParsingError, construct_grammar, ")")
        self.assertRaises(errors.GrammarParsingError, construct_grammar, "+(a:))")
        self.assertRaises(errors.GrammarParsingError, construct_grammar, "+(a:{)}")
        self.assertRaises(errors.GrammarParsingError, construct_grammar, "+ (a: 'a' )")
        self.assertRaises(errors.GrammarParsingError, construct_grammar, "+(a: 'a' sep)")
        self.assertRaises(errors.GrammarParsingError, construct_grammar, "+(a: 'a' sep {)")


    def test_parse_one_of_set(self):
        test_grammar = construct_grammar(r"""
            {
              { 'a' !"b" q'c'}
              {. q. u. }
              {  {$ .} }
            }
        """)

        self.assertIsInstance(test_grammar.sub_elements[0], elements.OneOfSet)
        self.assertIsInstance(test_grammar.sub_elements[0].sub_elements[0], elements.OneOfSet)
        self.assertIsInstance(test_grammar.sub_elements[0].sub_elements[1], elements.OneOfSet)
        self.assertIsInstance(test_grammar.sub_elements[0].sub_elements[2], elements.OneOfSet)

        se = test_grammar.sub_elements[0].sub_elements

        self.assertIsInstance(se[0].sub_elements[0], elements.StringLiteral)
        self.assertEqual(se[0].sub_elements[0].token_str, 'a')
        self.assertIsNone(se[0].sub_elements[0]._grammar_flags)

        self.assertIsInstance(se[0].sub_elements[1], elements.StringLiteral)
        self.assertEqual(se[0].sub_elements[1].token_str, 'b')
        self.assertEqual(se[0].sub_elements[1]._grammar_flags, {flags.NOT})

        self.assertIsInstance(se[0].sub_elements[2], elements.StringLiteral)
        self.assertEqual(se[0].sub_elements[2].token_str, 'c')
        self.assertEqual(se[0].sub_elements[2]._grammar_flags, {flags.QUOTED})


        self.assertIsInstance(se[1].sub_elements[0], elements.AnyString)
        self.assertIsNone(se[1].sub_elements[0]._grammar_flags)

        self.assertIsInstance(se[1].sub_elements[1], elements.AnyString)
        self.assertEqual(se[1].sub_elements[1]._grammar_flags, {flags.QUOTED})

        self.assertIsInstance(se[1].sub_elements[2], elements.AnyString)
        self.assertEqual(se[1].sub_elements[2]._grammar_flags, {flags.UNQUOTED})


        self.assertIsInstance(se[2].sub_elements[0], elements.OneOfSet)
        self.assertIsNone(se[2].sub_elements[0]._grammar_flags)

        self.assertIsInstance(se[2].sub_elements[0].sub_elements[0], elements.Newline)
        self.assertIsNone(se[2].sub_elements[0].sub_elements[0]._grammar_flags)

        self.assertIsInstance(se[2].sub_elements[0].sub_elements[1], elements.AnyString)
        self.assertIsNone(se[2].sub_elements[0].sub_elements[1]._grammar_flags)

        self.assertRaises(errors.GrammarParsingError, construct_grammar, "{a:.}")
        self.assertRaises(errors.GrammarParsingError, construct_grammar, "{a:")
        self.assertRaises(errors.GrammarParsingError, construct_grammar, "}")
        self.assertRaises(errors.GrammarParsingError, construct_grammar, "{)}")
        self.assertRaises(errors.GrammarParsingError, construct_grammar, "{'a' sep { 'b'}}")


    def test_parse_sub_grammar_definitions(self):
        test_grammar = construct_grammar(r"""
            def gramA { 'a' }
            def   gramB{'b'}
            def gramC { 'c'
            }
            def gramD
                {
                'd'}
            def gramE
                {'e'}
            def gramF{ 'f'    'f2'         }
            def
            gramG
            { 'g'
            }
            def gramH{ def gramI{ 'i' } 'h' gramI() }
            def gramJ {'j' def gramK{ 'k' }gramK()}

            def gramL{ 'l'
                def gramM{ 'm' }
                def gramN{ gramM() !'n' }
                gramN()
            }

            def gramO{
                def gramO2{ 'o' }
                def gramP{
                    def gramP2{ 'p' }
                    def gramQ{
                        def gramR{ 'r' }
                        gramO2()
                        gramP2()
                        'q'
                        gramR()
                    }
                    gramQ()
                }
                gramP()
            }

            def _{ }

            gramA()
             gramB
             ()
            gramC ()
            gramD        ()
            gramE()
            gramF()
            gramG()
            gramH()
            gramJ()
            gramL()
            gramO()
            _()
        """, allow_sub_grammar_definitions=True)

        self.assertListEqual([token.token_str for token in test_grammar.sub_elements], [
            'a', 'b', 'c', 'd', 'e', 'f', 'f2', 'g',
            'h', 'i', 'j', 'k', 'l', 'm', 'n',
            'o', 'p', 'q', 'r',
        ])

        self.assertRaises(errors.GrammarParsingError, construct_grammar, "def gramA{'a'}", allow_sub_grammar_definitions=False)
        self.assertRaises(errors.GrammarParsingError, construct_grammar, "def gram A{'a'}", allow_sub_grammar_definitions=True)
        self.assertRaises(errors.GrammarParsingError, construct_grammar, "def gram{A{'a'}", allow_sub_grammar_definitions=True)
        self.assertRaises(errors.GrammarParsingError, construct_grammar, "defgramA{'a'}", allow_sub_grammar_definitions=True)
        self.assertRaises(errors.GrammarParsingError, construct_grammar, "def gramA{ def gramB{ def gramC{ . } } } gramC()", allow_sub_grammar_definitions=True)
        self.assertRaises(errors.GrammarParsingError, construct_grammar, "def grammer", allow_sub_grammar_definitions=True)
        self.assertRaises(errors.GrammarParsingError, construct_grammar, "def gramA{ def gramB { 'b' } } gramB()", allow_sub_grammar_definitions=True)
        self.assertRaises(errors.GrammarParsingError, construct_grammar, "def gramA{ def gramB{ '' } } gramC()", allow_sub_grammar_definitions=True)



import inspect
import logging

from .grammar import flags, parse
from . import tokenizers
from .logger import LOGGER, TemporaryLogLevel

class Tokex(object):
    _grammar = None
    _tokenizer = None

    def __init__(self, input_grammar, allow_sub_grammar_definitions, tokenizer, default_flags=flags.DEFAULTS):
        self._grammar = parse.construct_grammar(input_grammar, allow_sub_grammar_definitions, default_flags)

        if inspect.isclass(tokenizer) and issubclass(tokenizer, tokenizers.TokexTokenizer):
            self._tokenizer = tokenizer()

        elif isinstance(tokenizer, tokenizers.TokexTokenizer):
            self._tokenizer = tokenizer

        else:
            raise Exception("Given tokenizer is not an instance of subclass of tokenizers.TokexTokenizer")


    # User-Level functions
    def match(self, input_string, match_entirety=True, debug=False):
        """
        Runs the loaded grammar against a string and returns the output if it matches the input string.

        Inputs: input_string   - The string to parse.
                match_entirety - A boolean, if True requires the entire string to be matched by the grammar.
                                if False, trailing tokens not matched by the grammar will not cause a match failure.
                debug          - A boolean, if True will set the debugging level to DEBUG for the duration of the match

        Outputs: A dictionary representing the output of parsing if the string matches the grammar, else None.
        """

        with TemporaryLogLevel(logging.DEBUG if debug else LOGGER.getEffectiveLevel()):
            tokens = self._tokenizer.tokenize(input_string)

            LOGGER.debug("Input Tokens:\n%s", tokens)

            match, end_idx, output = self._grammar.apply(tokens, 0)

            if match and (not match_entirety or end_idx == len(tokens)):
                return output[None] or {}

            return None



from ._base_element import BaseElement, BaseScopedElement
from .singular import AnyString, Newline, StringLiteral, RegexString
from .scoped import Grammar, NamedElement, IteratorDelimiter, ZeroOrOne, ZeroOrMore, OneOrMore, OneOfSet
from .sub_grammar import SubGrammarDefinition, SubGrammarUsage


# Dictionary which maps the first character of a grammar token to a grammar element class for all elements which
# can accept flags
FIRST_CHAR_VALID_FLAGS = {
    ".": AnyString,
    "~": RegexString,
    "'": StringLiteral,
    '"': StringLiteral,
}

FIRST_CHAR_ESCAPES = {
    '~',
    '"',
    "'"
}

__all__ = [
    "BaseElement",
    "BaseScopedElement",
    "AnyString",
    "Newline",
    "StringLiteral",
    "RegexString",
    "Grammar",
    "NamedElement",
    "IteratorDelimiter",
    "ZeroOrOne",
    "ZeroOrMore",
    "OneOrMore",
    "SubGrammarDefinition",
    "SubGrammarUsage"
]



import pdb
import pprint
import unittest

class TokexTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    def _debug(self, *args):
        pprint.pprint(*args)

    def _trace(self):
        pdb.set_trace()



import logging

LOGGER = logging.getLogger('tokex')

class TemporaryLogLevel(object):
    old_level = None

    def __init__(self, target_level):
        self.target_level = target_level

    def __enter__(self):
        self.old_level = LOGGER.getEffectiveLevel()
        LOGGER.setLevel(self.target_level)

    def __exit__(self, *_):
        LOGGER.setLevel(self.old_level)



from ._base_element import BaseScopedElement

class SubGrammarDefinition(BaseScopedElement):
    """
    Element for sub grammar scopes.
    Only used during grammar creation, these never actually makes it into a Grammar object
    """

    def human_readable_name(self):
        return "Sub Grammar def %s { ... }" % self.name

    def setup(self):
        self.sub_grammars = {}

        if self.token_str:
            # Find the name of the sub grammar; stripping off the def prefix
            self.name = self.name_re.search(self.token_str[3:]).group()


class SubGrammarUsage(BaseScopedElement):
    """
    Element for sub grammar replacements.
    Only used during grammar creation, these never actually makes it into a Grammar object
    """

    def human_readable_name(self):
        return "Sub Grammar Usage %s()" % self.name

    def setup(self):
        self.sub_grammars = {}

        if self.token_str:
            self.name = self.name_re.search(self.token_str).group()



import re

class TokexTokenizer(object):
    """
    Base class for Tokex tokenizers.  Uses re.findall & a collection of regular expressions to break up an
    input string into a sequence of tokens.

    Can be extended by subclassing this class & implementing a `tokenize` function or creating a custom
    list of tokenizer_regexes/tweaking tokenize_newlines

    Use of this function without subclassing will tokenize an input string by breaking up all occurrances of quotes
    into their own separate token; all sequences of alphanumeric tokens into their own separate token, and all
    concurrent non-alphanumeric space characters into their own token.
    """

    tokenizer_regexes = (
        r'"[^"]*"',
        r"'[^']*'",
        r"\b\w+\b",
        r"[^a-zA-Z0-9_ \t\n\r\f\v]+"
    )

    def __init__(self, tokenizer_regexes=None, tokenize_newlines=False, ignore_empty_lines=False):
        """
        Inputs: tokenizer_regexes  - Can be passed to provide a custom list of tokenizer regexes to parse
                                     an input string with.
                tokenize_newlines  - A boolean indicating whether newlines should be treated as tokens
                ignore_empty_lines - A boolean indicating whether we should skip over empty lines or not.
                                     Only has an effect if tokenize_newlines is passed and True
        """

        self.tokenize_newlines = tokenize_newlines
        self.ignore_empty_lines = ignore_empty_lines

        if tokenizer_regexes:
            self.tokenizer_regexes = tokenizer_regexes

        if self.tokenize_newlines:
            self.tokenizer_regexes = list(self.tokenizer_regexes) + [r"\n"]

    def tokenize(self, input_string):
        """
        Function which is called by tokex to break an input string into tokens, processed by tokex.

        Inputs: input_string - A string, to break into tokens.

        Outputs: A list of tokens from input_string.
        """

        tokens = re.findall(
            "(%s)" % "|".join(self.tokenizer_regexes),
            input_string,
            flags=re.MULTILINE
        )

        if self.tokenize_newlines and self.ignore_empty_lines:
            for idx in reversed(range(len(tokens))):
                if tokens[idx] == "\n" and ((idx > 0 and tokens[idx - 1] == "\n") or idx == 0):
                    tokens.pop(idx)

        return tokens


class NumericTokenizer(TokexTokenizer):
    """
    Tokenizers which keeps numeric values together as a single token
    """

    tokenizer_regexes = (
        r'"[^"]*"',
        r"'[^']*'",
        r"\S+",
    )



from ... import errors

from ._base_element import BaseScopedElement
from .singular import BaseSingular


class Grammar(BaseScopedElement):
    """ Named element which contains other grammar elements. """

    can_have_delimiter = False
    delimiter_grammar = None

    def setup(self):
        if self.token_str:
            self.name = self.name_re.search(self.token_str)
            if self.name:
                self.name = self.name.group()

    def human_readable_name(self):
        return "Named Section (%s: ...)" % self.name

    def _apply_sub_elements(self, string_tokens, idx):
        """
        Function which applies the sub elements of this element to the input tokens to see if they match

        Outputs: If our sub elements don't match the string tokens, False, None, None
                 Otherwise: A triple containing: {
                     match: A boolean depicting whether or not this construct matches the
                            iterable at the current position.
                     new_idx: The index this construct ceased matching upon the iterable, if match is True. Else None
                     output: A list of all returned dictionaries from the sub elements
                 )
        """

        outputs = {}

        for sub_element in self.sub_elements:
            match, idx, output = sub_element.apply(string_tokens, idx)

            if not match:
                return False, None, None

            if output is not None:
                outputs.update(output)

        return True, idx, outputs or None

    def _apply(self, string_tokens, idx):
        match, new_idx, output = self._apply_sub_elements(string_tokens, idx)

        if match:
            return True, new_idx, {self.name: output}

        return False, None, None


class NamedElement(Grammar):
    """ Named element which contains another singular element """

    def add_sub_element(self, sub_element):
        """
        Adds a sub element to this NamedElement.  This is used in places such as zero or more for example,
        where the zero or more element will contain one or more sub elements

        Inputs: sub_element - The element to add as a sub element of this element.
                              Should be a subclass of BaseGrammar
        """

        if self.sub_elements:
            raise errors.GrammarParsingError("%r cannot contain more than one element, already contains: %r" %
                                             (self, self.sub_elements[0]))

        if not isinstance(sub_element, BaseSingular):
            raise errors.GrammarParsingError("%r can only contain singular elements, not %r" % (self, sub_element))

        self.sub_elements.append(sub_element)

    def human_readable_name(self):
        return "Named Element <%s: ...>" % self.name

    def _apply(self, string_tokens, idx):
        if not self.sub_elements:
            return True, idx, None

        match, new_idx, _ = self.sub_elements[0].apply(string_tokens, idx)

        if match:
            return True, new_idx, {self.name: string_tokens[idx]}

        return False, None, None


class IteratorDelimiter(Grammar):
    """
    Element which can appear inside of another scoped container and causes the parent container to only
    match more than once if the delimiter matches between iterations
    """

    def setup(self):
        pass

    def human_readable_name(self):
        return "Iterator delimiter sep { ... }"

    def _apply(self, string_tokens, idx):
        match, idx, output = super(IteratorDelimiter, self)._apply(string_tokens, idx)

        if match and output is not None:
            output = output[None]

        return match, idx, output


class ZeroOrOne(Grammar):
    """ Element which can match a contained grammar zero or one times """

    def human_readable_name(self):
        return "Zero or One ?(%s: ...)" % self.name

    def _apply(self, string_tokens, idx):
        # If the index we're considering is beyond the end of our tokens we have nothing to match on.  However, since
        # we can match zero times, return True.  This allows gramars with trailing ZeroOrOne rules to match strings
        # which don't use them.
        if idx >= len(string_tokens):
            return True, idx, None

        match, new_idx, output = self._apply_sub_elements(string_tokens, idx)

        if match:
            if self.name:
                return True, new_idx, ({self.name: output} if new_idx > idx else None)
            return True, new_idx, output

        return True, idx, None


class ZeroOrMore(Grammar):
    """ Element which can match a contained grammar zero or more times """

    can_have_delimiter = True

    def human_readable_name(self):
        return "Zero or More *(%s: ...)" % self.name

    def _repeatedly_match(self, string_tokens, idx):
        match_count = 0
        current_idx = idx
        outputs = []
        while current_idx < len(string_tokens):
            new_idx = current_idx

            # If we're not processing the first match, check that any delimiter grammar we may have matches before
            # the next occurance of our grammar
            if match_count > 0 and self.delimiter_grammar is not None:
                match, new_idx, delimiter_output = self.delimiter_grammar.apply(string_tokens, new_idx)

                if not match:
                    break

                if delimiter_output:
                    outputs[-1].update(delimiter_output)

            # Try to match our sub elements
            match, new_idx, output = self._apply_sub_elements(string_tokens, new_idx)

            # If we don't match, or we do but we don't consume any tokens (ie we're stuck) exit the loop
            if not match or new_idx == current_idx:
                break

            match_count += 1
            outputs.append(output or None)
            current_idx = new_idx

        return match_count, current_idx, outputs

    def _apply(self, string_tokens, idx):
        # If the index we're considering is beyond the end of our tokens we have nothing to match on.  However, since
        # we can match zero times, return True.  This allows gramars with trailing ZeroOrMore rules to match strings
        # which don't use them.
        if idx >= len(string_tokens):
            return True, idx, None

        _, new_idx, outputs = self._repeatedly_match(string_tokens, idx)

        return True, new_idx, ({self.name: outputs} if new_idx > idx else None)


class OneOrMore(ZeroOrMore):
    """ Element which can match a contained grammar one or more times """

    can_have_delimiter = True

    def human_readable_name(self):
        return "One or More +(%s: ...)" % self.name

    def _apply(self, string_tokens, idx):
        match_count, idx, outputs = self._repeatedly_match(string_tokens, idx)

        if match_count > 0:
            return True, idx, {self.name: outputs}

        return False, None, None


class OneOfSet(Grammar):
    """ Element which can match any one of its contained grammars """

    def human_readable_name(self):
        return "One of Set {...}"

    def _apply(self, string_tokens, idx):
        for element in self.sub_elements:
            match, new_idx, output = element.apply(string_tokens, idx)
            if match:
                return True, new_idx, output

        return False, None, None



import argparse
import logging
import os
import sys
import unittest


# Globals
TEST_DIR = os.path.dirname(
    os.path.abspath(__file__)
)
PROJECT_ROOT = os.path.normpath(
    os.path.join(TEST_DIR, "..")
)

def run(verbosity=1):
    loader = unittest.TestLoader()
    suite = loader.discover(TEST_DIR)
    _runSuite(suite, verbosity)


def runSelective(testFiles, verbosity):
    loader = unittest.TestLoader()
    suite = loader.loadTestsFromNames(testFiles)
    _runSuite(suite, verbosity)


def _runSuite(testSuite, verbosity):
    unittest.TextTestRunner(verbosity=verbosity).run(testSuite)


def main():
    parser = argparse.ArgumentParser(description="Execute Tokex Unit Tests")
    parser.add_argument("testFiles", nargs="*")
    parser.add_argument("--verbosity", nargs="?", choices=['1', '2'], default=1)
    parser.add_argument("--debug", action="store_true")
    args = parser.parse_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)

    if len(args.testFiles):
        runSelective(args.testFiles, int(args.verbosity))
    else:
        run(int(args.verbosity))


if __name__ == '__main__':
    # Ensure we're running from the project root
    os.chdir(PROJECT_ROOT)
    sys.path.append(".")

    main()



import tokex
import _test_case

class TestGrammarParsing(_test_case.TokexTestCase):

    def test_parse_tokens(self):
        grammar = """
            'a' "b" s'c'
        """

        token_grammar = tokex.compile(grammar)

        self.assertIsNotNone(token_grammar.match('a b c'))
        self.assertIsNotNone(token_grammar.match('a b c '))
        self.assertIsNotNone(token_grammar.match(' a b c'))
        self.assertIsNotNone(token_grammar.match(' a  b  c '))

        self.assertIsNone(token_grammar.match("a b c d"))
        self.assertIsNone(token_grammar.match("c b c"))
        self.assertIsNone(token_grammar.match("a a c"))
        self.assertIsNone(token_grammar.match("a b b"))
        self.assertIsNone(token_grammar.match("a b C"))
        self.assertIsNone(token_grammar.match("a b"))
        self.assertIsNone(token_grammar.match("a"))
        self.assertIsNone(token_grammar.match(""))

        grammar = """
            . q. u.!'a_c'
        """
        regex_grammar = tokex.compile(grammar)

        self.assertIsNotNone(regex_grammar.match('anything "string" notstring nota_c'))
        self.assertIsNotNone(regex_grammar.match("anything 'string' notstring nota_c"))
        self.assertTrue(regex_grammar._grammar.apply(["anything", "'string'", "notstring", "nota\\_c"], 0)[0])

        self.assertIsNone(regex_grammar.match('anything "string" notstring a_c'))
        self.assertIsNone(regex_grammar.match('anything "string" "notstring" nota_c'))
        self.assertIsNone(regex_grammar.match('anything "string" \'notstring\' nota_c'))
        self.assertIsNone(regex_grammar.match('anything string notstring nota_c'))

        self.assertIsNone(tokex.match("s~caseSensitive~", "casesensitive"))
        self.assertIsNone(tokex.match("s~caseSensitive~", "CASESENSITIVE"))
        self.assertIsNotNone(tokex.match("i~caseSensitive~", "caseSensitive"))
        self.assertIsNotNone(tokex.match("i~caseSensitive~", "casesensitive"))
        self.assertIsNotNone(tokex.match("i~caseSensitive~", "CASESENSITIVE"))

        # Test newlines
        grammar = "'test' $ 'test'"
        newline_grammar = tokex.compile(grammar, tokenizer=tokex.tokenizers.TokexTokenizer(tokenize_newlines=True))
        self.assertIsNotNone(newline_grammar.match("test \n test"))
        self.assertIsNone(newline_grammar.match("test \\n test"))


    def test_parse_named_token(self):
        grammar = """
            <a1: 'a'>
            <a2: .>
            <a3: '>'>
            <a4: '<'>
            <a5: i'>'>
            <q:>
            <q2: >
            <a6: !'>'>
            <a7: 'q'>
            <a7: 'b'>
        """

        named_token_grammar = tokex.compile(grammar)

        self.assertDictEqual(named_token_grammar.match('a b > < > < q b'), {
            'a1': 'a',
            'a2': 'b',
            'a3': '>',
            'a4': '<',
            'a5': '>',
            'a6': '<',
            'a7': 'b'
        })

        self.assertIsNone(named_token_grammar.match(''))
        self.assertIsNone(named_token_grammar.match(' b > < > < q b'))
        self.assertIsNone(named_token_grammar.match('a b > < > < q'))
        self.assertIsNone(named_token_grammar.match('a b > < > > q b'))
        self.assertIsNone(named_token_grammar.match('b b > < > < q b'))

    def test_parse_named_grammar(self):
        grammar = """
            ( a:
                <a1: 'a'>
                <a2: 'b'>
                (a3: <a4: 'c'>)
            )
            (
             b:
                (c: (d: (e: <f: 'f'>) ) )
            )
            (c: 'c' 'c' 'c')
        """

        named_grammar_grammar = tokex.compile(grammar)

        self.assertDictEqual(named_grammar_grammar.match('a b c f c c c'), {
            'a': {
                'a1': 'a',
                'a2': 'b',
                'a3': {'a4': 'c'}
            },
            'b': {
                'c': {
                    'd': {
                        'e': {
                            'f': 'f'
                        }
                    }
                }
            },
            'c': None
        })

        self.assertIsNone(named_grammar_grammar.match('a b c'))
        self.assertIsNone(named_grammar_grammar.match('a b f'))
        self.assertIsNone(named_grammar_grammar.match('f'))
        self.assertIsNone(named_grammar_grammar.match('a c f'))
        self.assertIsNone(named_grammar_grammar.match('a a b c f'))
        self.assertIsNone(named_grammar_grammar.match(''))

    def test_parse_zero_or_one(self):
        grammar = """
            ?(a: 'a')
            ?(b:
                <bi: 'b'>
            )
            ?(c:
                (ci: <cii: 'c'>)
            )
            ?(d:
                ?(di:
                    ?(dii:
                        ?(diii:
                            <dv: 'd'>
                        )
                    )
                )
            )
            ?(e:
                ?(ei:
                    ?(eii:
                        ?(eiii:
                            (ev:<evi: 'e'>)
                        )
                    )
                )
            )
            ?(
                ?(
                    <outer: 'q'>
                )
                ?()
            )
        """

        zero_or_one_grammar = tokex.compile(grammar)

        self.assertDictEqual(zero_or_one_grammar.match('a b c d e q'), {
            'a': None,
            'b': {'bi': 'b'},
            'c': {'ci': {'cii': 'c'}},
            'd': {'di': {'dii': {'diii': {'dv': 'd'}}}},
            'e': {'ei': {'eii': {'eiii': {'ev': {'evi': 'e'}}}}},
            'outer': 'q'
        })

        self.assertDictEqual(zero_or_one_grammar.match('b c d e'), {
            'b': {'bi': 'b'},
            'c': {'ci': {'cii': 'c'}},
            'd': {'di': {'dii': {'diii': {'dv': 'd'}}}},
            'e': {'ei': {'eii': {'eiii': {'ev': {'evi': 'e'}}}}},
        })

        self.assertDictEqual(zero_or_one_grammar.match('c d e'), {
            'c': {'ci': {'cii': 'c'}},
            'd': {'di': {'dii': {'diii': {'dv': 'd'}}}},
            'e': {'ei': {'eii': {'eiii': {'ev': {'evi': 'e'}}}}},
        })

        self.assertDictEqual(zero_or_one_grammar.match('a b e'), {
            'a': None,
            'b': {'bi': 'b'},
            'e': {'ei': {'eii': {'eiii': {'ev': {'evi': 'e'}}}}},
        })

        self.assertDictEqual(zero_or_one_grammar.match('d e'), {
            'd': {'di': {'dii': {'diii': {'dv': 'd'}}}},
            'e': {'ei': {'eii': {'eiii': {'ev': {'evi': 'e'}}}}},
        })

        self.assertDictEqual(zero_or_one_grammar.match('e'), {
            'e': {'ei': {'eii': {'eiii': {'ev': {'evi': 'e'}}}}},
        })

        self.assertDictEqual(zero_or_one_grammar.match('q'), {
            'outer': 'q'
        })

        self.assertDictEqual(zero_or_one_grammar.match(''), {})

        self.assertIsNone(zero_or_one_grammar.match('f'))
        self.assertIsNone(zero_or_one_grammar.match('a b c d e f'))
        self.assertIsNone(zero_or_one_grammar.match('a b c g e f'))


    def test_parse_zero_or_more(self):
        grammar = """
            *(a:
                'a'
            )
            *(b:
                <bi: 'b'>
            )
            *(c:
                (c1:
                   <c2: 'c2'>
                   <c3: 'c3'>
                )
            )
            *(d:
                *(di:
                    *(dii:
                        *(diii:
                            <d: 'd'>
                        )
                    )
                )
            )
            *(e:
                *(ei:
                    *(eii:
                        *(eiii:
                            (e:<e: 'e'>)
                        )
                    )
                )
            )
        """

        zero_or_more_grammar = tokex.compile(grammar)

        self.assertDictEqual(zero_or_more_grammar.match('a b c2 c3 d e'), {
            'a': [None],
            'b': [{'bi': 'b'}],
            'c': [{'c1': {'c2': 'c2', 'c3': 'c3'}}],
            'd': [{'di': [{'dii': [{'diii': [{'d': 'd'}]}]}]}],
            'e': [{'ei': [{'eii': [{'eiii': [{'e': {'e': 'e'}}]}]}]}],
        })

        self.assertDictEqual(zero_or_more_grammar.match('a b d e'), {
            'a': [None],
            'b': [{'bi': 'b'}],
            'd': [{'di': [{'dii': [{'diii': [{'d': 'd'}]}]}]}],
            'e': [{'ei': [{'eii': [{'eiii': [{'e': {'e': 'e'}}]}]}]}],
        })

        self.assertDictEqual(zero_or_more_grammar.match('b c2 c3 d e'), {
            'b': [{'bi': 'b'}],
            'c': [{'c1': {'c2': 'c2', 'c3': 'c3'}}],
            'd': [{'di': [{'dii': [{'diii': [{'d': 'd'}]}]}]}],
            'e': [{'ei': [{'eii': [{'eiii': [{'e': {'e': 'e'}}]}]}]}],
        })

        self.assertDictEqual(zero_or_more_grammar.match('c2 c3 d e'), {
            'c': [{'c1': {'c2': 'c2', 'c3': 'c3'}}],
            'd': [{'di': [{'dii': [{'diii': [{'d': 'd'}]}]}]}],
            'e': [{'ei': [{'eii': [{'eiii': [{'e': {'e': 'e'}}]}]}]}],
        })

        self.assertDictEqual(zero_or_more_grammar.match('a b e'), {
            'a': [None],
            'b': [{'bi': 'b'}],
            'e': [{'ei': [{'eii': [{'eiii': [{'e': {'e': 'e'}}]}]}]}],
        })

        self.assertDictEqual(zero_or_more_grammar.match('d e'), {
            'd': [{'di': [{'dii': [{'diii': [{'d': 'd'}]}]}]}],
            'e': [{'ei': [{'eii': [{'eiii': [{'e': {'e': 'e'}}]}]}]}],
        })

        self.assertDictEqual(zero_or_more_grammar.match('e'), {
            'e': [{'ei': [{'eii': [{'eiii': [{'e': {'e': 'e'}}]}]}]}],
        })

        self.assertDictEqual(zero_or_more_grammar.match('a a a a a a a a a a b b b d d e e'), {
            'a': [None] * 10,
            'b': [{'bi': 'b'}, {'bi': 'b'}, {'bi': 'b'}],
            'd': [{'di': [{'dii': [{'diii': [{'d': 'd'}, {'d': 'd'}]}]}]}],
            'e': [{'ei': [{'eii': [{'eiii': [{'e': {'e': 'e'}}, {'e': {'e': 'e'}}]}]}]}],
        })

        self.assertDictEqual(zero_or_more_grammar.match('a a a a a a a a a a e e e e e e e'), {
            'a': [None] * 10,
            'e': [{'ei': [{'eii': [{'eiii': [{'e': {'e': 'e'}}, {'e': {'e': 'e'}}, {'e': {'e': 'e'}}, {'e': {'e': 'e'}},
                  {'e': {'e': 'e'}}, {'e': {'e': 'e'}}, {'e': {'e': 'e'}}]}]}]}],
        })

        self.assertDictEqual(zero_or_more_grammar.match('c2 c3 c2 c3 c2 c3'), {
            'c': [{'c1': {'c2': 'c2', 'c3': 'c3'}}, {'c1': {'c2': 'c2', 'c3': 'c3'}}, {'c1': {'c2': 'c2', 'c3': 'c3'}}],
        })

        self.assertDictEqual(zero_or_more_grammar.match(''), {})

        self.assertIsNone(zero_or_more_grammar.match('f'))
        self.assertIsNone(zero_or_more_grammar.match('a b c d e f'))
        self.assertIsNone(zero_or_more_grammar.match('a b c g e f'))


        grammar = """
            *(_:
                <a:'a'> sep{<b:'b'>}
            )
        """

        zero_or_more_grammar = tokex.compile(grammar)

        self.assertDictEqual(zero_or_more_grammar.match('a'), {'_': [{'a': 'a'}]})
        self.assertIsNone(zero_or_more_grammar.match('a b'))
        self.assertDictEqual(zero_or_more_grammar.match('a b a'), {'_': [{'a': 'a', 'b': 'b'}, {'a': 'a'}]})
        self.assertIsNone(zero_or_more_grammar.match('a b a b'))
        self.assertIsNone(zero_or_more_grammar.match('b'))
        self.assertIsNone(zero_or_more_grammar.match('a a'))


    def test_parse_one_or_more(self):
        grammar = """
            +(a:
                'a'
            )
            +(b:
                <bi: 'b'>
            )
            +(c:
                (c1:
                   <c2: 'c2'>
                   <c3: 'c3'>
                )
            )
            +(d:
                +(di:
                    +(dii:
                        +(diii:
                            <d: 'd'>
                        )
                    )
                )
            )
            +(e:
                +(ei:
                    +(eii:
                        +(eiii:
                            (e:<e: 'e'>)
                        )
                    )
                )
            )
        """

        one_or_more_grammar = tokex.compile(grammar)

        self.assertDictEqual(one_or_more_grammar.match('a b c2 c3 d e'), {
            'a': [None],
            'b': [{'bi': 'b'}],
            'c': [{'c1': {'c2': 'c2', 'c3': 'c3'}}],
            'd': [{'di': [{'dii': [{'diii': [{'d': 'd'}]}]}]}],
            'e': [{'ei': [{'eii': [{'eiii': [{'e': {'e': 'e'}}]}]}]}],
        })


        grammar = "+(_: <a:'a'> sep { <b:'b'> })"

        one_or_more_grammar = tokex.compile(grammar)

        self.assertDictEqual(one_or_more_grammar.match('a'), {'_': [{'a': 'a'}]})
        self.assertIsNone(one_or_more_grammar.match('a b'))
        self.assertDictEqual(one_or_more_grammar.match('a b a'), {'_': [{'a': 'a', 'b': 'b'}, {'a': 'a'}]})

        self.assertIsNone(one_or_more_grammar.match(''))
        self.assertIsNone(one_or_more_grammar.match('f'))
        self.assertIsNone(one_or_more_grammar.match('b'))
        self.assertIsNone(one_or_more_grammar.match('a a'))
        self.assertIsNone(one_or_more_grammar.match('a b'))
        self.assertIsNone(one_or_more_grammar.match('a b a b'))
        self.assertIsNone(one_or_more_grammar.match('b'))
        self.assertIsNone(one_or_more_grammar.match('a a'))


        grammar = """
            +(_:
                <a:'a'> sep {<b:'b'> }
            )
        """

        one_or_more_grammar = tokex.compile(grammar)

        self.assertDictEqual(one_or_more_grammar.match('a'), {'_': [{'a': 'a'}]})
        self.assertIsNone(one_or_more_grammar.match('a b'))
        self.assertDictEqual(one_or_more_grammar.match('a b a'), {'_': [{'a': 'a', 'b': 'b'}, {'a': 'a'}]})
        self.assertIsNone(one_or_more_grammar.match('a b a b'))
        self.assertIsNone(one_or_more_grammar.match('a a'))
        self.assertIsNone(one_or_more_grammar.match('b'))


    def test_parse_one_of_set(self):
        grammar = """
            {
                <a:'a'>
                (b: <b1:'b1'> 'b2')
                {
                    <c:'c'>
                    'd'
                    'e'
                }
            }
        """

        one_of_set_grammar = tokex.compile(grammar)

        self.assertDictEqual(one_of_set_grammar.match('a'), {'a': 'a'})
        self.assertDictEqual(one_of_set_grammar.match('b1 b2'), {'b': {'b1': 'b1'}})
        self.assertDictEqual(one_of_set_grammar.match('c'), {'c': 'c'})
        self.assertDictEqual(one_of_set_grammar.match('d'), {})
        self.assertDictEqual(one_of_set_grammar.match('e'), {})

        self.assertIsNone(one_of_set_grammar.match(''))
        self.assertIsNone(one_of_set_grammar.match('b3'))
        self.assertIsNone(one_of_set_grammar.match('b1'))
        self.assertIsNone(one_of_set_grammar.match('b2'))

        grammar = """
            *(_:
                {
                    <a:'a'>
                    (b: <b1:'b1'> <b2:'b2'>)
                    {
                        <c:'c'>
                        'd'
                        'e'
                    }
                }
            )
        """

        one_of_set_grammar = tokex.compile(grammar)

        self.assertDictEqual(one_of_set_grammar.match('a a a a c c c d d d d d e e e e e e b1 b2 a a a b1 b2'), {
            '_': [
                {'a': 'a'}, {'a': 'a'}, {'a': 'a'}, {'a': 'a'},
                {'c': 'c'}, {'c': 'c'}, {'c': 'c'},
                None, None, None, None, None, None, None, None, None, None, None,
                {'b': {'b1': 'b1', 'b2': 'b2'}},
                {'a': 'a'}, {'a': 'a'}, {'a': 'a'},
                {'b': {'b1': 'b1', 'b2': 'b2'}}
            ]
        })

        self.assertDictEqual(one_of_set_grammar.match(''), {})
        self.assertDictEqual(one_of_set_grammar.match('e'), {'_': [None]})



from setuptools import setup, find_packages

with open("README.md") as fd:
    long_description=fd.read()

setup(
    name='tokex',
    version='2.0.2',
    description="String tokenizing and parsing library",
    long_description=long_description,
    long_description_content_type="text/markdown",

    author="Warren Spencer",
    author_email="warrenspencer27@gmail.com",
    url="https://github.com/warrenspe/tokex",
    keywords=["tokex", "tokenize", "parse"],

    packages=find_packages(),

    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Software Development :: Libraries",
        "Topic :: Text Processing"
    ]
)



class TokexError(Exception):
    """ Base class for all tokex-specific errors """


class GrammarParsingError(TokexError):
    """ Base class for errors which occur while parsing an input tokex grammar """


class UnknownGrammarTokenError(GrammarParsingError):
    """ Error thrown when we are unable to tokenize a user-defined grammar. """

    def __init__(self, token):
        super(UnknownGrammarTokenError, self).__init__()

        self.token = token

    def __str__(self):
        return "Unknown grammar token: %s" % self.token


class InvalidGrammarTokenFlagsError(GrammarParsingError):
    """ Error thrown when a grammar token in a user-defined grammar has an invalid flag """

    def __init__(self, flags, element):
        super(InvalidGrammarTokenFlagsError, self).__init__()

        self.flags = flags
        self.element = element

    def __str__(self):
        add_s = ""
        if len(self.flags) > 1:
            add_s = "s"
        return "Invalid flag%s given to %r: %s.  Valid flags are: %s" % (
            add_s,
            self.element,
            "".join(self.flags),
            self.element.valid_flags
        )


class MismatchedBracketsError(GrammarParsingError):
    """
    Error thrown when opening/closing brackets for a given element are mismatched
    For example, opening for a Zero or more, and closing for a one of set
    """

    def __init__(self, bracket, current_scope=None):
        super(MismatchedBracketsError, self).__init__()

        self.bracket = bracket
        self.current_scope = current_scope

    def __str__(self):
        if self.current_scope:
            return "Mismatched bracket given: %s for scoped element: %r" % (self.bracket, self.current_scope)

        return "Mismatched bracket given: %s" % self.bracket


class SubGrammarError(GrammarParsingError):
    """ Raised when a user-defined grammar has invalid sub grammars within it """

    def __init__(self, msg):
        super(SubGrammarError, self).__init__(self)
        self.msg = msg

    def __str__(self):
        return self.msg



"""
File containing utilities to parse a user-defined grammar string into a grammar tree
"""

import re

from .. import errors
from . import elements
from . import flags

def tokenize_grammar(grammar_string):
    """ Function which accepts a grammar string and returns an iterable of tokens """

    # Function to create regex's for given flags
    flags_re_string = lambda f: "[%s]*" % "".join(f)

    name_re_str = elements.BaseScopedElement.name_re_str

    pattern = "|".join((
        # Defined Sub Grammar open
        r"def\s+%s\s*\{" % name_re_str,
        # Defined Sub Grammar usage
        r"%s\s*\(\s*\)" % name_re_str,
        # Iterator Delimiter Open
        r"sep\s*\{",
        # ZeroOrMore, OneOrMore, ZeroOrOne, and Named Grammar open
        r"[*+?]?\(\s*%s\s*:" % name_re_str,
        # ZeroOrOne Unnamed Grammar open
        r"\?\(",
        # Named Token open
        r"<\s*%s?\s*:" % name_re_str,
        # One of Set open
        r"\{",
        # All Regex
        r"%s\." % flags_re_string(elements.AnyString.valid_flags),
        # Regex
        r"%s~(?:[^\\~]*(?:\\.)*)*~" % flags_re_string(elements.RegexString.valid_flags),
        # Literal String
        #r"%s'.*?(?<!\\)'" % flags_re_string(elements.StringLiteral.valid_flags),
        r"%s'(?:[^\\']*(?:\\.)*)*'" % flags_re_string(elements.StringLiteral.valid_flags),
        # Literal String
        #r'%s".*?(?<!\\)"' % flags_re_string(elements.StringLiteral.valid_flags),
        r'%s"(?:[^\\"]*(?:\\.)*)*"' % flags_re_string(elements.StringLiteral.valid_flags),
        # Newline Token
        r"\$",

        # Closers
        # Named Token close
        r">",
        # Flow & Named Grammar close
        r"\)",
        # Sub Grammar, Iterator Delimiter, & One of Set close
        r"\}",

        # Special non-token-class tokens to match
        # Comments
        r"(?P<_comment_>#[^\n]*(\n|$))",
        # Final fallback - nontoken
        r"(?P<_nontoken_>\S+)"
    ))

    matched_tokens = []
    all_flags_re = re.compile("[%s]+" % "".join((getattr(flags, flag) for flag in flags.__all__)))
    escape_re = re.compile(r"\\(.)")

    for match in re.finditer(pattern, grammar_string, re.I):
        if match.groupdict().get('_nontoken_'):
            raise errors.UnknownGrammarTokenError(match.groupdict().get('_nontoken_'))

        elif match.groupdict().get('_comment_'):
            continue

        matched_token = match.group()

        # Check for flags on the token
        token_flags = None
        if matched_token[:3].lower() not in ("def", "sep") and matched_token[:-2] != "()":
            token_flags = all_flags_re.match(matched_token)
            if token_flags:
                token_flags = set(token_flags.group())
                matched_token = matched_token[len(token_flags):]

                # Ensure the flags set on the element are valid
                element = elements.FIRST_CHAR_VALID_FLAGS[matched_token[0]]
                invalid_flags = token_flags.difference(element.valid_flags)
                if invalid_flags:
                    raise errors.InvalidGrammarTokenFlagsError(invalid_flags, element)

        # Handle escapes in the token
        if matched_token[0] in elements.FIRST_CHAR_ESCAPES:
            matched_token = matched_token[0] + escape_re.sub(r"\1", matched_token[1:-1]) + matched_token[-1]

        matched_tokens.append({
            "token": matched_token,
            "flags": token_flags
        })

    return matched_tokens


def construct_grammar(grammar_string, allow_sub_grammar_definitions=False, default_flags=flags.DEFAULTS):
    """
    Function which accepts a user-defined grammar string and returns an instance of Grammar representing it.

    Inputs: grammar_string - The user-defined grammar string representing the grammar we should construct.
            allow_sub_grammar_definitions - A boolean, indicating whether or not we should process sub grammars
                                            See the README for more information on what sub grammars are used for
                                            and the dangers of allowing them when parsing untrusted third-party grammars
            default_flags - Can be passed as a set of flags, which will set the defaults for all elements in the grammar

    Outputs: An instance of a Grammar class which can be used to parse input strings.
    """

    grammar_tokens = tokenize_grammar(grammar_string)

    # The grammar stack; opening tokens add a new item to the stack, closing tokens pop one off
    # Pre-populated with an outer-most grammar that will be returned from this function
    grammar_stack = [elements.Grammar(default_flags=default_flags)]

    # A stack of sub_grammars; used to handle nested sub_grammar definitions
    # We prepopulate it with an outer-most subgrammar to handle global sub-grammar definitions
    sub_grammar_stack = [elements.SubGrammarDefinition(default_flags=default_flags)]

    for token_idx, token_dict in enumerate(grammar_tokens):
        if not grammar_stack:
            raise errors.MismatchedBracketsError(grammar_tokens[token_idx - 1])

        token_flags = token_dict["flags"]
        token = token_dict["token"]

        # Openers
        if token == "{":
            element = elements.OneOfSet(token, token_flags, default_flags)
            grammar_stack[-1].add_sub_element(element)
            grammar_stack.append(element)

        elif token[:2] == "*(":
            element = elements.ZeroOrMore(token, token_flags, default_flags)
            grammar_stack[-1].add_sub_element(element)
            grammar_stack.append(element)

        elif token[:2] == "+(":
            element = elements.OneOrMore(token, token_flags, default_flags)
            grammar_stack[-1].add_sub_element(element)
            grammar_stack.append(element)

        elif token[:2] == "?(":
            element = elements.ZeroOrOne(token, token_flags, default_flags)
            grammar_stack[-1].add_sub_element(element)
            grammar_stack.append(element)

        elif token[0] == "(":
            element = elements.Grammar(token, token_flags, default_flags)
            grammar_stack[-1].add_sub_element(element)
            grammar_stack.append(element)

        elif token[0] == "<":
            element = elements.NamedElement(token, token_flags, default_flags)
            grammar_stack[-1].add_sub_element(element)
            grammar_stack.append(element)

        elif token[:3].lower() == "sep":
            element = elements.IteratorDelimiter(token, token_flags, default_flags)
            if grammar_stack[-1].delimiter_grammar:
                raise errors.GrammarParsingError("Multiple %r defined for %r" % (element, grammar_stack[-1]))

            if not getattr(grammar_stack[-1], "can_have_delimiter", False):
                raise errors.GrammarParsingError("Cannot add iterator delimiter to %r" % grammar_stack[-1])

            grammar_stack[-1].delimiter_grammar = element
            grammar_stack.append(element)

        # Closers
        elif token == "}":
            if isinstance(grammar_stack[-1],
                          (elements.SubGrammarDefinition, elements.IteratorDelimiter, elements.OneOfSet)):
                if isinstance(grammar_stack[-1], elements.SubGrammarDefinition):
                    new_sub_grammar = sub_grammar_stack.pop()
                    sub_grammar_stack[-1].sub_grammars[new_sub_grammar.name] = new_sub_grammar

                grammar_stack.pop()

            else:
                raise errors.MismatchedBracketsError(token, grammar_stack[-1])

        elif token == ")":
            if isinstance(grammar_stack[-1],
                          (elements.ZeroOrMore, elements.ZeroOrOne, elements.OneOrMore, elements.Grammar)):
                grammar_stack.pop()

            else:
                raise errors.MismatchedBracketsError(token, grammar_stack[-1])

        elif token == ">":
            if isinstance(grammar_stack[-1], (elements.NamedElement)):
                grammar_stack.pop()

            else:
                raise errors.MismatchedBracketsError(token, grammar_stack[-1])

        # Singular tokens
        elif token[0] in ("'", '"'):
            grammar_stack[-1].add_sub_element(elements.StringLiteral(token, token_flags, default_flags))

        elif token[0] == "~":
            grammar_stack[-1].add_sub_element(elements.RegexString(token, token_flags, default_flags))

        elif token[0] in ("'", '"'):
            grammar_stack[-1].add_sub_element(elements.StringLiteral(token, token_flags, default_flags))

        elif token == "$":
            grammar_stack[-1].add_sub_element(elements.Newline(token, token_flags, default_flags))

        elif token == ".":
            grammar_stack[-1].add_sub_element(elements.AnyString(token, token_flags, default_flags))

        # Sub Grammar open
        elif token[:3].lower() == "def":
            if not allow_sub_grammar_definitions:
                raise errors.SubGrammarError("Cannot define a Sub-grammar when allow_sub_grammar_definitions is False")

            # Only allow definition of a new subgrammar within the global scope and other subgrammars
            for element in grammar_stack[1:]:
                if not isinstance(element, elements.SubGrammarDefinition):
                    raise errors.SubGrammarError("Sub grammars cannot be defined within a %r element" % element)

            element = elements.SubGrammarDefinition(token, token_flags, default_flags)
            grammar_stack.append(element)
            sub_grammar_stack.append(element)

        # Sub Grammar Usage
        elif token[-1] == ")":
            # Find the referenced sub_grammar
            sub_grammar_name = elements.SubGrammarUsage(token, token_flags, default_flags).name

            for parent_sub_grammar in reversed(sub_grammar_stack):
                if sub_grammar_name in parent_sub_grammar.sub_grammars:
                    for sub_element in parent_sub_grammar.sub_grammars[sub_grammar_name].sub_elements:
                        grammar_stack[-1].add_sub_element(sub_element)

                    break

            else:
                raise errors.SubGrammarError("Cannot find named sub grammar: %s" % sub_grammar_name)


        else:
            raise errors.GrammarParsingError("Unknown token: %s" % repr(token))

        if not grammar_stack:
            raise errors.MismatchedBracketsError(token)

    if len(grammar_stack) > 1:
        raise errors.MismatchedBracketsError(grammar_stack[-1].token_str)

    return grammar_stack[0]



from .logger import LOGGER as logger
from .functions import compile, match
from . import tokenizers, errors
from .grammar import flags

__all__ = [
    "compile",
    "match",
    "tokenizers",
    "errors",
    "flags",
    "logger"
]
